// Error handler system adapted to FiveWin
// ErrSysW.prg

#include "error.ch"
#include "FiveWin.ch"
*#include "PPPWIN.ch"

external _fwGenError   // Link FiveWin generic Error Objects Generator

#define NTRIM(n)    ( LTrim( Str( n ) ) )
#command QUIT => ( PostQuitMessage( 0 ), __Quit() ) // suggested by Enrico

STATIC cCtlFile := "qcsctl.ppp"  // Updated by ap's main proc for ProPlus.

/*************
*	ErrorSys()
*
*	Note:  automatically executes at startup
*/
proc ErrorSys()
MEMVAR errorsys
* 1 line of code added to lock out decompilers:
        errorsys=9876543210
* ^     ^ continue with orig code\/:

    ErrorBlock( { | e | ErrorDialog( e ) } )
return

proc ErrorLink()
return


/*************
*   ErrorDialog()
*/
static function ErrorDialog( e ) // -> logical or quits App.

   local oDlg, oLbx, oFont
   local lRet    // if lRet == nil -> default action: QUIT
   local n, j, cMessage, aStack := {}
   local oSay, hLogo
   local nButtons  := 1
   local cErrorLog := ""
   local aVersions := GetVersion()
   local aTasks := {} //   := GetTasks()  // This change to prevent HUGE memory leak
   local aRDDs, nTarget, uValue
   local oOldError  // should be bOldError!!
   local cRelation
   local lIsWinNT := IsWinNT()
    // added by RWS...
   local Stopq
*   memvar C_QPATH

*   IF(EMPTY(C_QPATH), C_QPATH := "\QCS360X",)

   // RWS - per FiveTech newsgroup advice 8/02...
   // by default, division by zero yields zero - (Moved up to prevent HUGE memory leak.)
   if ( e:genCode == EG_ZERODIV )
      return 0
   end
   
aTasks := GetTasks()  // Moved down to prevent HUGE memory leak by this func,
                      // if called prior to divide-by-zero check. RWS 8/02

   cErrorLog += "Application" + CRLF
   cErrorLog += "===========" + CRLF
   cErrorLog += "   Path and name: " + GetModuleFileName( GetInstance() ) + CRLF
   cErrorLog += "   Size: " + Transform( FSize( GetModuleFileName( ;
                GetInstance() ) ), "9,999,999 bytes" ) + CRLF
   cErrorLog += "   Max files handles permited: ( SetHandleCount() ) " + ;
                Str( SetHandleCount(), 3 ) + CRLF
   cErrorLog += "   Error ocurred at: " + ;
                DToC( Date() ) + ", " + Time() + CRLF

   // Error object analysis
   cMessage   = "   Error description: " + ErrorMessage( e ) + CRLF
   cErrorLog += cMessage

*----------
   // RWS - per Patrick Mast advice on FiveTech newsgroup 12/02 to eliminate
   // LOW MEMORY error with XP machines...

   IF "Memory Low"$cMessage
      SysRefresh()
      Memory(-1)
      return  .f.
   ENDIF
*----------

   if ValType( e:Args ) == "A"
      cErrorLog += "   Args:" + CRLF
      for n = 1 to Len( e:Args )
         cErrorLog += "     [" + Str( n, 4 ) + "] = " + ValType( e:Args[ n ] ) + ;
                      "   " + cValToChar( e:Args[ 1 ] ) + CRLF
      next
   endif


   cErrorLog += CRLF + "Stack Calls" + CRLF
   cErrorLog += "===========" + CRLF
      n := 2    // we don't disscard any info again !
      while ( n < 74 )
          if ! Empty(ProcName( n ) )
             AAdd( aStack, "   Called from " + Trim( ProcName( n ) ) + ;
                           "(" + NTRIM( ProcLine( n ) ) + ")" )
             cErrorLog += ATail( aStack ) + CRLF
          endif
          n++
    end

   cErrorLog += CRLF + "System" + CRLF
   cErrorLog += "======" + CRLF

   cErrorLog += "   CPU type: " + GetCPU() + CRLF
   // cErrorLog += "   CPU type: " + { "386", "486", "Pentium" }[ CPUType() - 2 ] + CRLF
   cErrorLog += "   Hardware memory: " + cValToChar( nExtMem() + 768 ) + ;
                " (" + cValToChar( Int( ( nExtMem() + 768 ) / 1024 ) ) + ;
                " megas)" + CRLF + CRLF
   cErrorLog += "   Free System resources: % " + AllTrim( Str( GetFreeSystemResources( 0 ) ) ) + CRLF + ;
                "        GDI    resources: % " + AllTrim( Str( GetFreeSystemResources( 1 ) ) ) + CRLF + ;
                "        User   resources: % " + AllTrim( Str( GetFreeSystemResources( 2 ) ) ) + CRLF + CRLF
   cErrorLog += "   Windows and MsDos versions: " + ;
                Str( aVersions[ 1 ], 2 ) + "." + ;
                Str( aVersions[ 2 ], 2 ) + ", " + ;
                Str( aVersions[ 3 ], 2 ) + "." + ;
                Str( aVersions[ 4 ], 2 ) + CRLF
   cErrorLog += "   Windows total applications running: " + Str( GetNumTasks(), 3 ) + CRLF
   for n = 1 to Len( aTasks )
      cErrorLog += "    " + Str( n, 3 ) + " " + aTasks[ n ] + CRLF
   next


/* Original position...
   // by default, division by zero yields zero
   if ( e:genCode == EG_ZERODIV )
      return 0
   end
*/
   // for network open error, set NETERR() and subsystem default
   if ( e:genCode == EG_OPEN .and. ;
      ( e:osCode == 32 .or. e:osCode == 5 ) .and. ;
        e:canDefault )
      NetErr( .t. )
      return .f.       // Warning: Exiting!
   end

   // for lock error during APPEND BLANK, set NETERR() and subsystem default
   if ( e:genCode == EG_APPENDLOCK .and. e:canDefault )
      NetErr( .t. )
      return .f.       // OJO SALIDA
   endif


   if ProcName( 7 ) == "ERRORDIALO"   // recursive error !!!
      CLOSE DATABASES
      SET RESOURCES TO
      ResAllFree()
      ErrorLevel( 1 )
*      CLEAR MEMORY
      QUIT
   endif


   // Warning!!! Keep here this code !!! Or we will be consuming GDI as
   // we don't generate the error but we were generating the bitmap

*   hLogo = FWBitMap()

   if e:canRetry
      nButtons++
   endif

   if e:canDefault
      nButtons++
   endif

   cErrorLog += CRLF + "Variables in use" + CRLF + "================" + CRLF
   cErrorLog += "   Procedure     Type   Value" + CRLF
   cErrorLog += "   ==========================" + CRLF

   n := 2    // we don't disscard any info again !
   while ( n < 74 )

       if ! Empty( ProcName( n ) )
          cErrorLog += "   " + Trim( ProcName( n ) ) + CRLF
          for j = 1 to ParamCount( n )
             cErrorLog += "     Param " + Str( j, 3 ) + ":    " + ;
                          ValType( GetParam( n, j ) ) + ;
                          "    " + cGetInfo( GetParam( n, j ) ) + CRLF
          next
          for j = 1 to LocalCount( n )
             cErrorLog += "     Local " + Str( j, 3 ) + ":    " + ;
                          ValType( GetLocal( n, j ) ) + ;
                          "    " + cGetInfo( GetLocal( n, j ) ) + CRLF
          next
       endif

       n++
   end

   cErrorLog += CRLF + "Linked RDDs" + CRLF + "===========" + CRLF
   aRDDs = RddList( 1 )
   for n = 1 to Len( aRDDs )
      cErrorLog += "   " + aRDDs[ n ] + CRLF
   next

   cErrorLog += CRLF + "DataBases in use" + CRLF + "================" + CRLF
   for n = 1 to 255
      if ! Empty( Alias( n ) )
         cErrorLog += CRLF + Str( n, 3 ) + ": " + If( Select() == n,"=> ", "   " ) + ;
                      PadR( Alias( n ), 15 ) + Space( 20 ) + "RddName: " + ;
                      ( Alias( n ) )->( RddName() ) + CRLF
         cErrorLog += "     ==============================" + CRLF
         cErrorLog += "     RecNo    RecCount    BOF   EOF" + CRLF
         cErrorLog += "    " + Transform( ( Alias( n ) )->( RecNo() ), "99999" ) + ;
                      "      " + Transform( ( Alias( n ) )->( RecCount() ), "99999" ) + ;
                      "      " + cValToChar( ( Alias( n ) )->( BoF() ) ) + ;
                      "   " + cValToChar( ( Alias( n ) )->( EoF() ) ) + CRLF + CRLF
         cErrorLog += "     Indexes in use " + Space( 23 ) + "TagName" + CRLF
         for j = 1 to 15
            if ! Empty( ( Alias( n ) )->( IndexKey( j ) ) )
               cErrorLog += Space( 8 ) + ;
                            If( ( Alias( n ) )->( IndexOrd() ) == j, "=> ", "   " ) + ;
                            PadR( ( Alias( n ) )->( IndexKey( j ) ), 35 ) + ;
                            ( Alias( n ) )->( OrdName( j ) ) + ;
                            CRLF
            endif
         next
         cErrorLog += CRLF + "     Relations in use" + CRLF
         for j = 1 to 8
            if ! Empty( ( nTarget := ( Alias( n ) )->( DbRSelect( j ) ) ) )
               cErrorLog += Space( 8 ) + Str( j ) + ": " + ;
                            "TO " + ( Alias( n ) )->( DbRelation( j ) ) + ;
                            " INTO " + Alias( nTarget ) + CRLF
               // uValue = ( Alias( n ) )->( DbRelation( j ) )
               // cErrorLog += cValToChar( &( uValue ) ) + CRLF
            endif
         next
      endif
   next

   n = 1
   cErrorLog += CRLF + "Classes in use:" + CRLF
   cErrorLog += "===============" + CRLF
   while ! Empty( __ClassName( n ) )
      cErrorLog += "   " + Str( n, 3 ) + " " + __ClassName( n++ ) + CRLF
   end

   cErrorLog += CRLF + "Memory Analysis" + CRLF
   cErrorLog +=        "===============" + CRLF
   cErrorLog += "   Static memory:" + CRLF
   cErrorLog += "      data segment: 64k" + CRLF
   cErrorLog += "      Initial size:       " + ;
                LTrim( Str( nInitDSSize() ) ) + ;
                " bytes  (SYMP=" + LTrim( Str( nSymPSize() ) ) + ;
                ", Stack=" + LTrim( Str( nStackSize() ) ) + ;
                ", Heap=" + LTrim( Str( nHeapSize() ) ) + ")" + CRLF
   cErrorLog += "      Clipper Stack:      " + ;
                LTrim( Str( 65535 - ( nStatics() * 14 ) - nInitDSSize() ) ) + ;
                " bytes" + CRLF
   cErrorLog += "      " + LTrim( Str( nStatics() ) ) + " Static variables: " + ;
                LTrim( Str( nStatics() * 14 ) ) + " bytes" + CRLF + CRLF

   cErrorLog += "   Dynamic memory consume:" + CRLF
   cErrorLog += "      Actual  Value: " + Str( MemUsed() ) + " bytes" + CRLF
   cErrorLog += "      Highest Value: " + Str( MemMax() ) + " bytes" + CRLF
   // nSymNames() no longer returns a real value! 15/April/97
   /*
   cErrorLog += "   SYMBOLS segment" + CRLF
   cErrorLog += "      " + LTrim( Str( nSymNames() ) ) + " SymbolNames:   " + ;
                LTrim( Str( nSymNames() * 16 ) ) + " bytes"
   */

   // Generates a file with an Error Log

   BEGIN SEQUENCE
      oOldError = ErrorBlock( { || DoBreak() } )
      MemoWrit( "Error.log", cErrorLog )
      IF !lIsDir("c:\transfer")
         lMkDir("c:\transfer")
      ENDIF
      // Make sure the error.log is saved locally as well, in case it gets
      // overwritten on the server by an error from another workstation before
      // it is seen.
      MemoWrit( "c:\transfer\Error.log", cErrorLog )
   END SEQUENCE
   ErrorBlock( oOldError )

   DEFINE FONT oFont NAME "Ms Sans Serif" SIZE 0, -10
*xx
   DEFINE DIALOG oDlg ;
      SIZE 320, 250 + If( lIsWinNT, 50, 0 ) ;
      TITLE OemToAnsi( "Program Error Report" ) ;
      FONT oFont

   @ 0, 0 SAY oSay PROMPT OemToAnsi( cMessage ) ;
      BORDER OF oDlg FONT oFont SIZE 150, 14

   oSay:nStyle   = nOR( oSay:nStyle, 128 )   // SS_NOPREFIX
   oSay:nTop     =   4
   oSay:nLeft    =   5 // 22
   oSay:nBottom  =  17 // 25
   oSay:nRight   = 156 //148

   @ 21, 6 SAY "( Email c:\transfer\ERROR.LOG file to roger@leadersoft.com )" OF oDlg FONT oFont PIXEL
   @ 30, 6 SAY "&Stack List - note 1st item with non-zero number in parenthesis:" OF oDlg FONT oFont PIXEL

   n = aStack[ 1 ]

   @ 45, 7 LISTBOX oLbx VAR n ITEMS aStack OF oDlg ;
      SIZE 145, 62 + If( lIsWinNT, 18, 0 ) PIXEL

   if nButtons == 1 .or. nButtons == 3
      @ 108 + If( lIsWinNT, 24, 0 ), 60 BUTTON "&Quit" OF oDlg ACTION oDlg:End() ;
         SIZE 30, 12 PIXEL FONT oFont DEFAULT
   else
      @ 108 + If( lIsWinNT, 24, 0 ), 37 BUTTON "&Quit" OF oDlg ACTION oDlg:End() ;
         SIZE 30, 12 PIXEL FONT oFont
   endif

   if e:CanRetry
      @ 108 + If( lIsWinNT, 24, 0 ), If( nButtons == 2, 82, 13 ) BUTTON "&Retry" ;
         OF oDlg ACTION ( lRet  := .t., oDlg:End() ) ;
         SIZE 30, 12 FONT oFont PIXEL
   endif

   if e:CanDefault
      @ 108 + If( lIsWinNT, 24, 0 ), 108 BUTTON "&Default"  OF oDlg ;
         ACTION ( lRet  := .f., oDlg:End() ) ;
         SIZE 30, 12 FONT oFont PIXEL
   endif

   ACTIVATE DIALOG oDlg CENTERED ;
      RESIZE16

   oFont:End()

   if lRet == nil .or. ( !LWRunning() .and. lRet )
      CLOSE DATABASES
      SET RESOURCES TO
      ErrorLevel( 1 )

      // Regarding ProPlus network simultaneous users control:
      // it is important to call the following function to log this
      // workstation out before EVERY exit of the program --
*xh      IF _pp_lanact() = 1  // .AND. _pp_lanuse() > 0
*         _pp_initcf( RTRIM(C_QPATH) +"\"+ cCtlFile )
*         _pp_landec( 1 )      
*      ENDIF

*      STOPQ := RTRIM(C_QPATH) + "\STOPQCS.MEM"
*      IF FILE((STOPQ))
*        ERASE (Stopq)
*      ENDIF

*      CLEAR MEMORY


      ResAllFree()
      SET 3DLOOK OFF

      QUIT              // must be QUIT !!!
   endif

return lRet

//----------------------------------------------------------------------------//

static function DoBreak()

   BREAK

return nil

//----------------------------------------------------------------------------//

static func ErrorMessage( e )

	// start error message
    local cMessage := if( empty( e:OsCode ), ;
                          if( e:severity > ES_WARNING, "Error ", "Warning " ),;
                          "(DOS Error " + NTRIM(e:osCode) + ") " )

	// add subsystem name if available
    cMessage += if( ValType( e:SubSystem ) == "C",;
                    e:SubSystem()                ,;
                    "???" )

	// add subsystem's error code if available
    cMessage += if( ValType( e:SubCode ) == "N",;
                    "/" + NTRIM( e:SubCode )   ,;
                    "/???" )
	// add error description if available
  if ( ValType( e:Description ) == "C" )
        cMessage += "  " + e:Description
	end

	// add either filename or operation
    cMessage += if( ! Empty( e:FileName ),;
                    ": " + e:FileName   ,;
                    if( !Empty( e:Operation ),;
                        ": " + e:Operation   ,;
                        "" ) )
return cMessage

//----------------------------------------------------------------------------//
// returns extended info for a certain variable type

static function cGetInfo( uVal )

   local cType := ValType( uVal )

   do case
      case cType == "C"
           return '"' + cValToChar( uVal ) + '"'

      case cType == "O"
           return "Class: " + uVal:ClassName()

      case cType == "A"
           return "Len: " + Str( Len( uVal ), 4 )

      otherwise
           return cValToChar( uVal )
   endcase

return nil

//--------------------------------------------------------------------------/

  // A work-around due to Microsoft change in Win XP...
function GetFreeSystemResources()

return 90

//--------------------------------------------------------------------------/
   // Called from main proc to set name of ProPlus control file.
PROCEDURE CntrlUpd(cCtl)

  cCtlFile := cCtl

RETURN
*------------------