#include "FiveWin.ch"
#include "report.ch"

#define UND_SINGLE        1
#define UND_DOUBLE        2

//----------------------------------------------------------------------------//

CLASS TRColumn

     DATA   oReport
     DATA   aData, aTitle, aPicture
     DATA   bDataFont, bTitleFont, bTotalFont, bTotalExpr, bColor
     DATA   cTotalPict
     DATA   nWidth, nDataHeight, nTitleHeight, nTotal, nCol, nSize, nPad,;
            nPen, nOrder, nCalCol, nCurLine
     DATA   lTotal, lShadow, lGrid, lTotalExpr, lUnderline, lDobleUnd, lTextUnd,;
            lSeparator
     DATA   Cargo

     METHOD New( aTitle, nCol, aData, nSize, aPicture,;
                 bFont, lTotal, bTotalExpr, cFmt, lShadow, lGrid,;
                 oReport ) CONSTRUCTOR

     METHOD Stabilize()

     METHOD SayTitle( nRow, nCol, nLine )
     METHOD SayData( nRow, nCol, nLine )
     METHOD SayTotal( nRow, nCol )
     METHOD Separator( nDataLine, nRow )

     METHOD SayMemo(cMemo, cText, oReport) // Added by R. Seiler 11/99
     METHOD GetTop(cText, nFor, nLines)    //  "    "   " for printing memo fields.
     METHOD Underline( lTextLength, lDouble, nDataLine, nRow )
     METHOD TitleHeight( nLine )

ENDCLASS

//----------------------------------------------------------------------------//

METHOD New( aTitle     ,;
            nCol       ,;
            aData      ,;
            nSize      ,;
            aPicture   ,;
            bFont      ,;
            lTotal     ,;
            bTotalExpr ,;
            cFmt       ,;
            lShadow    ,;
            lGrid      ,;
            nPen       ,;
            oReport     )  CLASS TRColumn

     DEFAULT aTitle      := {{|| ""} }                   ,;
             aData       := {{|| ""} }                   ,;
             nSize       := 0                            ,;
             aPicture    := {""}                         ,;
             nCol        := 0                            ,;
             bFont       := {|| 1 }                      ,;
             nPen        := 1                            ,;
             lTotal      := .F.                          ,;
             lShadow     := .F.                          ,;
             lGrid       := .F.                          ,;
             cFmt        := iif(Valtype(Eval(aData[1]))=="N",;
                                "RIGHT", "LEFT")

     ::aTitle     = aTitle
     ::aData      = aData
     ::nSize      = nSize
     ::aPicture   = aPicture
     ::bDataFont  = bFont
     ::bTitleFont = {|| 1 }
     ::bTotalFont = bFont
     ::lTotal     = lTotal
     ::lShadow    = lShadow
     ::lGrid      = lGrid
     ::lUnderline = .F.
     ::lDobleUnd  = .F.
     ::lSeparator = .T.
     ::lTextUnd   = .F.
     ::nPen       = nPen
     ::nTotal     = 0
     ::oReport    = oReport
     ::bTotalExpr = bTotalExpr
     ::nCol       = nCol
     ::cTotalPict = iif(len(aPicture)>0, aPicture[1], "")
     ::nWidth     = 0
     ::nCalCol    = 0
     ::nOrder     = 0

     ::lTotalExpr = !(bTotalExpr == NIL)

     IF cFmt == "LEFT"
          ::nPad = RPT_LEFT
     ELSEIF cFmt == "RIGHT"
          ::nPad = RPT_RIGHT
     ELSEIF cFmt == "CENTER" .OR. cFmt == "CENTERED"
          ::nPad = RPT_CENTER
     ELSE
          ::nPad = RPT_LEFT
     ENDIF

     DO WHILE len(::aPicture) < len(::aData)
          Aadd(::aPicture,::cTotalPict)
     ENDDO

RETURN Self

//----------------------------------------------------------------------------//

METHOD Stabilize(nOrder) CLASS TRColumn

     LOCAL nFor, nWidth, nLen

     ::nDataHeight  := ::oReport:oDevice:GetTextHeight(::oReport:cCharPattern,;
                       ::oReport:aFont[eval(::bDataFont)])
     ::nTitleHeight := ::oReport:oDevice:GetTextHeight(::oReport:cCharPattern,;
                       ::oReport:aFont[eval(::bTitleFont)])

     IF !empty(::nSize)
          ::nWidth := ::oReport:oDevice:GetTextWidth(Replicate(::oReport:cCharPattern,::nSize),;
                       ::oReport:aFont[eval(::bDataFont)])
          RETU NIL
     ENDIF


     FOR nFor := 1 TO len(::aData)
          nLen      := len(cValtoChar(Transform(eval(::aData[nFor]),;
                           ::aPicture[nFor])))
          nWidth    := ::oReport:oDevice:GetTextWidth(Replicate(::oReport:cCharPattern,nLen),;
                       ::oReport:aFont[eval(::bDataFont)])
          ::nWidth  := Max(::nWidth,nWidth)
     NEXT

     FOR nFor := 1 TO len(::aTitle)
          nLen      := len(eval(::aTitle[nFor]))
          nWidth    := ::oReport:oDevice:GetTextWidth(Replicate(::oReport:cCharPattern,nLen),;
                       ::oReport:aFont[eval(::bTitleFont)])
          ::nWidth  := Max(::nWidth,nWidth)
     NEXT

     ::nOrder := nOrder

RETURN NIL

//----------------------------------------------------------------------------//

METHOD SayTitle(nRow, nCol, nLine)  CLASS TRColumn

     LOCAL oFont
     LOCAL cTitle
     LOCAL nWidth, nFont

     DEFAULT nRow  := 0      ,;
             nCol  := ::nCol ,;
             nLine := 1

     IF nLine > len(::aTitle)
          RETU NIL
     ENDIF

     cTitle := eval(::aTitle[nLine])
     nFont  := eval(::bTitleFont)
     oFont  := ::oReport:aFont[nFont]

     ::nCurLine := nLine

     ::oReport:oDevice:Say(nRow, nCol, cTitle, oFont, ::nWidth,;
                           ::oReport:aClrText[nFont],,::nPad-1)

RETURN NIL


//----------------------------------------------------------------------------//

METHOD SayData(nRow, nCol, nLine)  CLASS TRColumn

     LOCAL oFont
     LOCAL cText
     LOCAL nWidth, nFont, lMemo   // lMemo added by RWS

     DEFAULT nRow  := 0      ,;
             nCol  := ::nCol ,;
             nLine := 1

     IF nLine > len(::aData)
          RETU NIL
     ENDIF

     ::nCurLine := nLine

IF VALTYPE(eval(::aData[nLine])) == "M"   // Memo printing section added
   lMemo := .t.                           // by R. Seiler 11/99 with call
  ELSE                                    // to METHOD SayMemo()
   lMemo := .f.
ENDIF

IF lMemo
   ::SayMemo(eval(::aData[nLine]),"",::oReport)
  ELSE

     nFont  := eval(::bDataFont)
     oFont  := ::oReport:aFont[nFont]
     cText  := Transform(eval(::aData[nLine]),::aPicture[nLine])
     nWidth := ::oReport:oDevice:GetTextWidth(cText, oFont)

/* From earlier RWS version...
     DO CASE
          CASE ::nPad == RPT_RIGHT
               nCol := nCol+::nWidth-nWidth

          CASE ::nPad == RPT_CENTER
               nCol := nCol+Int(::nWidth/2)-Int(nWidth/2)
     ENDCASE
*/
     ::oReport:oDevice:Say(nRow, nCol, cText, oFont, ::nWidth,;
                           ::oReport:aClrText[nFont],,::nPad-1)

      IF ::lUnderline
          ::Underline(::lTextUnd        ,;
                      ::lDobleUnd       ,;
                      nLine             ,;
                      nRow+::nDataHeight)
      ENDIF

ENDIF // lMemo

RETURN NIL

//----------------------------------------------------------------------------//

METHOD SayTotal(nRow, nCol)  CLASS TRColumn

     LOCAL oFont
     LOCAL cText
     LOCAL nFont

     DEFAULT nRow  := 0      ,;
             nCol  := ::nCol

     IF !::lTotal
          RETU NIL
     ENDIF

     nFont := eval(::bTotalFont)
     oFont := ::oReport:aFont[nFont]
     cText := Transform(::nTotal,::cTotalPict)

     ::oReport:oDevice:Say(nRow, nCol, cText, oFont, ::nWidth, ::oReport:aClrText[nFont],,::nPad-1)

RETURN NIL

//----------------------------------------------------------------------------//

METHOD TitleHeight( nLine ) CLASS TRColumn

     LOCAL oFontT, oFontD
     LOCAL cText, cChar
     LOCAL nFor, nWidth, nLen, nFontT, nFontD, nLenData

     IF !empty(::nWidth)
          RETU NIL
     ENDIF

     nFontT := eval(::bTitleFont)
     nFontD := eval(::bDataFont)
     oFontT := ::oReport:aFont[nFontT]
     oFontD := ::oReport:aFont[nFontD]
     cChar := ::oReport:cCharPattern

     ::nDataHeight  := len(::aData) * ::oReport:oDevice:GetTextHeight(cChar, oFontD)
     ::nTitleHeight := len(::aTitle) * ::oReport:oDevice:GetTextHeight(cChar, oFontT)

     nLenData := len(::aData)

     FOR nFor := 1 TO nLenData
          cText    := Transform(eval(::aData[nLine]),::aPicture[nLine])
          nLen     := len(cText)
          nWidth   := ::oReport:oDevice:GetTextWidth(Replicate(cChar,nLen), oFontD)
          ::nWidth := Max(::nWidth,nWidth)
     NEXT

RETURN NIL

//----------------------------------------------------------------------------//

METHOD Underline(lTextUnd, lDobleUnd, nLine, nRow)  CLASS TRColumn

     LOCAL nLeft, nRight, nWidth

     DEFAULT lTextUnd  := ::lTextUnd  ,;
             lDobleUnd := ::lDobleUnd ,;
             nLine     := 1           ,;
             nRow      := ::oReport:nRow

     DO CASE
     CASE !lTextUnd .AND. ::lGrid
          nLeft  := ::nCalCol  + Int(::oReport:nSeparator/2)
          nRight := nLeft + ::nWidth + Int(::oReport:nSeparator/2)
     CASE lTextUnd
          nWidth := ::oReport:oDevice:GetTextWidth(alltrim(;
                      Transform(eval(::aData[nLine]),::aPicture[nLine])),;
                      ::oReport:aFont[eval(::bDataFont)])
          DO CASE
               CASE ::nPad == RPT_RIGHT
                    nLeft := ::nCalCol+::nWidth-nWidth
               CASE ::nPad == RPT_CENTER
                    nLeft := ::nCalCol+Int(::nWidth/2)-Int(nWidth/2)
               OTHERWISE
                    nLeft := ::nCalCol
          ENDCASE
          nRight := nLeft + nWidth
     OTHERWISE
          nLeft  := ::nCalCol
          nRight := nLeft + ::nWidth
     ENDCASE


     IF lDobleUnd
          nWidth := ::oReport:aPen[::nPen]:nWidth
          ::oReport:oDevice:Line(nRow-nWidth,;
                                 nLeft,;
                                 nRow-nWidth,;
                                 nRight,;
                                 ::oReport:aPen[::nPen])

          ::oReport:oDevice:Line(nRow+nWidth,;
                                 nLeft,;
                                 nRow+nWidth,;
                                 nRight,;
                                 ::oReport:aPen[::nPen])
     ELSE
          ::oReport:oDevice:Line(nRow,;
                                 nLeft,;
                                 nRow,;
                                 nRight,;
                                 ::oReport:aPen[::nPen])
     ENDIF

RETURN NIL

//----------------------------------------------------------------------------//

METHOD Separator(nRow, lForced) CLASS TRColumn

   LOCAL nLeft, nRight, nSep

   DEFAULT nRow    := ::oReport:nRow,;
           lForced := .f.

   IF !::lSeparator .and. !lForced
      RETURN NIL
   ENDIF

   nSep   := iif( ::lGrid, Int(::oReport:nSeparator/2), 0)
   nLeft  := ::nCalCol  - nSep
   nRight := nLeft + ::nWidth + (nSep*2)

   ::oReport:oDevice:Line(nRow,;
                          nLeft,;
                          nRow,;
                          nRight,;
                          ::oReport:aPen[::nPen])

RETURN NIL
*------------------------
// Print memo field...  (Added by R. Seiler, 11/99)
// Can't use ::oReport here - for unknown reason it can't be found by
// this method, so have to pass oReport to it as a parameter.
METHOD SayMemo(cMemo,cText,oReport)  CLASS TRColumn
     LOCAL cLine
     LOCAL nPageln
     LOCAL nLines := 0,nFor := 0,nCols := 0,lTop := .f.

   SysRefresh()
   CursorWait()  

   nCols := LEN(oReport:aColumns)
   
   IF EMPTY(cText) .AND. !EMPTY(cMemo)
      cText := cMemo
   ENDIF
     nLines := MlCount(cText,74) // Original text has up to 74 chars. per line.
     nPageln := 0
*     oReport:BackLine(1)     // Not needed here. If used, will cause printing
                              // of subsequent pages to get out of synch by
                              // pushing the start of printing down 1 line
                              // per page.
     FOR nFor := 1 TO nLines

          cLine := MemoLine(cText, 74, nFor)

          // The next 4 lines handle situation where new page follows a
          // short page with less than 60 lines. This is
          // triggered by "Page No."  Only works if "Page No." appears on
          // top left of page.
          IF SUBSTR(LTRIM(cLine),1,8) == "Page No." .AND. nFor > 3
             oReport:EndPage()
             nPageln := 0
          ENDIF
          IF oReport:nRow < 325  // To prevent memo from overprinting 1st row.
             oReport:Endline()
             lTop := .t.         //     "        "         "
          ENDIF
          IF nFor = 1 .AND. !lTop  // lTop prevents NewLine() from adding 
             oReport:NewLine()     // extra space if using Endline().
          ENDIF
          lTop := .f.
          oReport:StartLine()
          IF nCols > 3
             oReport:Say(2,cLine)
            ELSE
             oReport:Say(1,cLine)
          ENDIF
          oReport:EndLine()

             // The following section checks to see if we've reached the
             // bottom of the page (60 lines). If so, it then calls
             // GetTop() to eliminate empty spaces in the text file between
             // the bottom of the outgoing page and the top line of the
             // next page.  The intent of this code is to get rid of the
             // empty lines that most report generators insert between pages
             // when output is to a text file -- otherwise our page printing
             // would get out of synch with the formating of the original
             // report. This approach only works where printing (including the
             // page number and other page heading data) is supposed to appear
             // at the same place at the top of every page.  
          nPageln := nPageln + 1
          IF nPageln = 60
             nFor := ::GetTop(cText,nFor,nLines)
             nPageln := 0
          ENDIF

          SysRefresh()
          CursorWait()
     NEXT

*    oReport:Newline()  // Only activate this line if you need to have the
                        // program eject another page upon completing printing.

RETURN NIL

*----------------------

METHOD GetTop(cText,nFor,nLines)  CLASS TRColumn
LOCAL Test := .T.,cLine
DO WHILE Test = .T. .AND. nFor <= nLines
   nFor := nFor + 1
   cLine := MemoLine(cText, 76, nFor)
   TEST := EMPTY(cLine)
   SysRefresh()   // Inserted to allow Windows to catch up with housekeeping
                  // that is held up by the "DO WHILE."
ENDDO
nFor := nFor - 1
RETURN nFor
*----------------------
