/*----------------------------------------------------------------------------//
  TstRpt23.prg  - demo TRept23() and ToText() - enhancements to Treport system.
       By Roger Seiler - roger@leadersoft.com

       Rept23.prg was first written in 1995, was then greatly enhanced over
       the years, and now has been converted to OOP. Both oop and no-oop
       versions are provided.

       TRept23/Rept23 provides a way to use TReport to print out a report
       previously created and saved to disk in a text file. It enables use
       of multiple fonts and graphics - lines, boxes, shading and bitmaps.

       However, the report can alternatively be passed to TRept23/Rept23 as
       a memvar from the report-creating program module without saving it
       as a text file.

       This was orignally conceived as a way of being able to use reports
       in Windows (via FW) originally written in Clipper for DOS, either in
       the traditional @ SAY way or via report writers like R&R (with Clipper
       Code Generator) or Bandit. But the Rept23 approach has grown into 
       an important extension to TReport's capabilities as a report engine.

       It's companion, ToText(), provides an easy way to quickly convert
       Clipper @ SAY style report code into a Windows report printed via
       TRept23/Rept23. So ToText() serves as an intermediate translator between
       old Clipper report code and Windows printing via TRept23/Rept23.

       Detailed commentary on how to use these programs is included at the
       bottom of each prg.

       Anyone who wants to see an example of how to convert code from
       "procedural" to OOP can compare Rept23.prg to TRept23.prg.

//----------------------------------------------------------------------------*/

#include "FiveWin.ch"

memvar oWnd
static nPrnLinAdj := 0,cRefDir := "",aSet := {},cIniPrefx := "MyApp"

//----------------------------------------------------------------------------//

function Main()

   local oIco, oBar, nAt := 0, oMsg, cMsg, oFont

   public oWnd

   cRefDir := CurDir()
   IF AT("\", cRefDir) > 0
      cRefDir := SUBSTR(cRefDir, AT("\", cRefDir), LEN(cRefDir) )
   ENDIF

   DEFINE ICON oIco FILE "fivewin.ico"
   DEFINE FONT oFont NAME "MS SANS SERIF" SIZE 0,-10  

   DEFINE WINDOW oWnd FROM 1, 1 TO 32, 85 ;
      TITLE "Testing TRept23() Sample Reports" ;
      MENU  BuildMenu() ;
      COLOR "N/W" ;
      ICON oIco

      oWnd:oFont := oFont  

   DEFINE BUTTONBAR oBar _3D SIZE 33, 33 OF oWnd

   DEFINE BUTTON OF oBar FILENAME "edit.bmp" NOBORDER ;
      TOOLTIP "View Readme.txt" ;
      MESSAGE "View Readme.txt with explanations of TRept23 and ToText." ;
      ACTION ( WinExec("EDIT.COM Readme.txt",3) )

   DEFINE BUTTON OF oBar FILENAME "exit.bmp" NOBORDER ;
      TOOLTIP "Exit Program" ;
      MESSAGE "Bug out." ;
      ACTION oWnd:End()

*  DEFINE MESSAGE oMsg OF oWnd PROMPT cMsg CLOCK DATE KEYBOARD
   SET MESSAGE OF oWnd TO "TstRpt23 message bar" ;
      CENTERED DATE KEYBOARD 2007

   ACTIVATE WINDOW oWnd ;
     ON INIT ( PrintAdj("MyApp") )  
               // Automatically enter print alignment settings in .ini file
               // the first time ever that the app is launched by workstn.

return nil

//----------------------------------------------------------------------------//

function BuildMenu()

local oMenu,oRept23,aSpool,aFonts,cGet := SPACE(20),lOkay := .f.

aSpool := {} // Contains names of spool files for large printouts over 64K.
             // Spool names added to array in RRGENRC rr_prline().

aFonts := {;
          {"TIMES NEW ROMAN", 0,-11, ""},;
          {"TIMES NEW ROMAN", 0,-11, "BOLD"},;
          {"TIMES NEW ROMAN", 0,-11, "ITALIC"},;
          {"", 0,0, ""},;
          {"ARIAL", 0,-14, "BOLD"},;
          {"ARIAL", 0,-20, "BOLD"};
          }


   MENU oMenu 2007
      MENUITEM "&Reports"
      MENU
         MENUITEM "&View Readme.txt file" ;
            MESSAGE "See report textfile with special codes for multi-fonts." ;
            ACTION ( WinExec("EDIT.COM Readme.txt",3) )

         SEPARATOR

         MENUITEM "&Mixed fonts report - See textfile" ;
            MESSAGE "See report textfile with special codes for multi-fonts." ;
            ACTION ( WinExec("EDIT.COM SamplPA1.txt",3) )

         MENUITEM "M&ixed fonts report - Print" ;
            MESSAGE "Mixed fonts report - no graphics." ;
            ACTION ( oRept23 := TRept23():New("SamplPA1.txt","",aFonts,aSpool,.t. ), oRept23:Activate(), oRept23 := nil )

         SEPARATOR

         MENUITEM "&Report With Graphics - See textfile"  ;
            MESSAGE "See report textfile with special codes for graphics." ;
            ACTION ( WinExec("EDIT.COM SamplPA2.txt",3) )

         MENUITEM "R&eport With Graphics - Print"  ;
            MESSAGE "Mixed fonts report - with graphic boxes, shading and lines." ;
            ACTION ( oRept23 := TRept23():New("SamplPA2.txt","",aFonts,aSpool,.t. ), oRept23:Activate(), oRept23 := nil )

         SEPARATOR
/*
         MENUITEM "&Test MsgGet"  ;
            MESSAGE "Test." ;
            ACTION (lOkay := MsgGet("Describe Subdirectory","Description:",@cGet) )
*/
         MENUITEM "E&xit"  ;
            MESSAGE "Exit application." ;
            ACTION ( oWnd:End() )

      ENDMENU

      MENUITEM "&Utilities"
      MENU
         MENUITEM "&Print Position Adjustments" ;
            MESSAGE "Adjust printing alignment." ;
            ACTION ( PrnLen(nPrnLinAdj,oWnd,"Q") )

      ENDMENU
   ENDMENU

return oMenu

//----------------------------------------------------------------------------//
*--------------
function GetRefDir() ; RETURN cRefDir  // Pass filewide static to calling proc
function GetAset() ; RETURN aSet  // Pass filewide static to calling proc
function GetIniNam() ; RETURN cIniPrefx
FUNCTION SetPLadj(nPLadj); nPrnLinAdj := nPLadj ; RETURN nil

procedure AppSys  // Alaska XBase++ requirement

return 